unit HashTable;

interface

uses
  Classes;

type
  THashNode = class (TObject)
    Parent : THashNode;
    Data : TObject;
    constructor Create;
    destructor Destroy; override;
  end;

  THashArray = Array[0..36] of THashNode;

  THashSubNode = class (THashNode)
    HashArray : THashArray;
    destructor Destroy; override;
  end;

  THashTable = class (TObject)
  private
    Root : THashSubNode;
  public
    constructor Create;
    destructor Destroy; override;
    function LookupOrAppend(Value : String) : THashNode; virtual;
    procedure CopyToStringList(SL : TStringList);
  end;

  TListHashTable = class (THashTable)
    function LookupOrAppend(Value : String) : THashNode; override;
  end;

  TListHashNode = class (TList)
    destructor Destroy; override;
  end;

function Hash(C : Char) : Byte;
function UnHash(B : Byte) : Char;

implementation

uses
  SysUtils;

function Hash(C : Char) : Byte;
begin
  if C in ['0'..'9'] then
    Result := Ord(C) - Ord('0')
  else if C in ['a'..'z'] then
    Result := Ord(C) - Ord('a') + 10
  else if C in ['A'..'Z'] then
    Result := Ord(C) - Ord('A') + 10
  else if C = '_' then
    Result := 36
  else
    raise Exception.CreateFmt('Invalid symbol in hash string ("%s" #%d)',[C, Ord(C)]);
end;

function UnHash(B : Byte) : Char;
begin
  if B < 10 then
    Result := Char(Ord('0') + B)
  else if B < 36 then
    Result := Char(Ord('A') + B - 10)
  else
    Result := '_';
end;

constructor THashNode.Create;
begin
  inherited Create;
  Data := Nil;
end;

destructor THashNode.Destroy;
begin
  Data.Free;
  Data := Nil;
  inherited Destroy;
end;

destructor THashSubNode.Destroy;
var
  I : Integer;
begin
  for I := 0 to 36 do
    HashArray[I].Free;
  inherited Destroy;
end;

constructor THashTable.Create;
begin
  inherited Create;
  Root := THashSubNode.Create;
end;

destructor THashTable.Destroy;
begin
  Root.Free;
  inherited Destroy;
end;

function THashTable.LookupOrAppend(Value : String) : THashNode;
var
  Ptr : THashSubNode;
  Work : THashSubNode;
  Next : THashNode;
  HashValue : Byte;
  Last : Boolean;
  ValueLength : Integer;
  I : Integer;
begin
  Ptr := Root;
  ValueLength := Length(Value);
  for I := 1 to ValueLength do
    begin
      Last := I = ValueLength;
      HashValue := Hash(Value[I]);
      Next := Ptr.HashArray[HashValue];
      if Next = Nil then
        begin
          if Last then
            begin
              Result := THashNode.Create;
              Result.Parent := Ptr;
              Ptr.HashArray[HashValue] := Result;
              Exit;
            end
          else
            begin
              Next := THashSubNode.Create;
              THashNode(Next).Parent := Ptr;
              Ptr.HashArray[HashValue] := Next;
              Ptr := THashSubNode(Next);
            end;
        end
      else
        begin
          if Last then
            begin
              Result := THashNode(Next);
              Exit;
            end
          else if not (Next is THashSubNode) then
            begin
              Work := THashSubNode.Create;
              Work.Parent := THashNode(Next).Parent;
              Work.Data := THashNode(Next).Data;
              Ptr.HashArray[HashValue] := Work;
              THashNode(Next).Data := Nil;
              Next.Free;
              Ptr := Work;
            end
          else
            Ptr := THashSubNode(Next);
        end;
    end;
  Result := Ptr;
end;

procedure CheckNode(SL : TStringList; Str : String; Node : THashNode);
var
  I : Integer;
  Obj : THashnode;
begin
  if Node.Data <> Nil then
    SL.AddObject(Str, Node.Data);

  if Node is THashSubNode then with THashSubNode(Node) do
    begin
      For I := 0 to 36 do
        begin
          Obj := HashArray[I];
          if Obj <> Nil then
            CheckNode(SL, Str + UnHash(I), Obj);
        end;
    end;
end;

procedure THashTable.CopyToStringList(SL : TStringList);
begin
  SL.Clear;
  CheckNode(SL, '', Root);
end;

function TListHashTable.LookupOrAppend(Value : String) : THashNode;
var
  Node : THashNode;
  List : TListHashNode;
begin
  Node := inherited LookupOrAppend(Value);
  if Node.Data = Nil then
    begin
      List := TListHashNode.Create;
      Node.Data := List;
    end
  else
    List := TListHashNode(Node.Data);

  Result := THashNode.Create;
  List.Add(Result);
end;

destructor TListHashNode.Destroy;
var
  I : Integer;
begin
  For I := 0 to Count-1 do
    THashNode(Items[I]).Free;
  inherited Destroy;
end;

end.
