unit XMLParser;

interface

uses
  Classes,
  SysUtils;

type
  // Abstract class
  TXMLNode = class(TObject)
  private
    MyTagName: String;
    MyData: String;
    MyChildren: TList;
    procedure LoadInternals(IntTagName: String; AStream: TStream);
    function GetItem(AIndex: Integer): TXMLNode;
    function GetCount: Integer;
    function GetNode(AStream: TStream) : TXMLNode;
  public
    constructor Create(TagName: String; Value: String); overload;
    constructor CreateRoot(AStream: TStream); overload;
    constructor Create(TagName: String; AStream: TStream; var LineNum: Integer); overload;
    destructor Destroy; override;
    function GetChildNode(ATagName: String) : TXMLNode;
    function GetAttributeValue(AName: String) : String;
    property TagName: String read MyTagName;
    property Data: String read MyData;
    property Items[Index: Integer]: TXMLNode read GetItem; default;
    property Count: Integer read GetCount;
  end;

  TXMLParser = class(TObject)
  private
    MyFileName: String;
    MyRoot: TXMLNode;
  public
    constructor Create(const AFileName: String);
    destructor Destroy; override;
    function GetNode(Name: String) : TXMLNode;
  end;

implementation

const
  ErrorUnexpectedEOF = 'Unexpected end of file at line %d';
  ErrorUnexpectedChar = 'Unexpected character "%s" at line %d';
  ErrorUnexpectedWhitespace = 'Unexpected whitspace at line %d';
  ErrorTagMismatch = 'End tag at line %d (%s) does not match start tag (%s)';
  ErrorNodeNotFound = 'Node "%s" not found';

constructor TXMLNode.Create(TagName: String; Value: String);
begin
  inherited Create;
  MyTagName := TagName;
  MyData := Value;
end;

constructor TXMLNode.Create(TagName: String; AStream: TStream; var LineNum: Integer);
type
  TState = (
    LookingForAttribNameOrEnd,
    LookingForEquals,
    LookingForValue,
    ReadingValue,
    ReadingAttribName,
    LookingForEnd
    );
var
  State: TState;
  AttribName: String;
  AttribValue: String;
  DoubleQuotes: Boolean;
  C: Char;
begin
  inherited Create;
  MyTagName := TagName;
  DoubleQuotes := false;
  State := LookingForAttribNameOrEnd;

  while True do
    begin
      // Read the next character
      if(AStream.Read(C,1) <> 1) then
        C := #00
      else
        C := UpCase(C);

      case C of
        #10: Inc(LineNum);
        #13: begin end; // Do nothing
      else
        case State of
          LookingForAttribNameOrEnd:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              ' ': begin end; // Do nothing
              '/': State := LookingForEnd;
              '>':
                begin
                  LoadInternals(TagName, AStream);
                  Exit;
                end;
              'A'..'Z','0'..'9','_':
                begin
                  AStream.Seek(-1, soFromCurrent);
                  State := ReadingAttribName;
                  AttribName := '';
                end;
            end;
          ReadingAttribName:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              'A'..'Z','0'..'9','_':
                AttribName := AttribName + C;
              else
                begin
                  AStream.Seek(-1, soFromCurrent);
                  State := LookingForEquals;
                end;
            end;
          LookingForEquals:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              ' ': begin end; // Nothing
              '=': State := LookingForValue;
            else
              raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
          LookingForValue:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              '"':
                begin
                  DoubleQuotes := True;
                  State := ReadingValue;
                  AttribValue := '';
                end;
              '''':
                begin
                  DoubleQuotes := false;
                  State := ReadingValue;
                  AttribValue := '';
                end;
            else
              raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
          ReadingValue:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              '"':
                if DoubleQuotes then
                  begin
                    // Add AttribName, AttribValue
                    If MyChildren = NIL then
                      MyChildren := TList.Create;
                    MyChildren.Add(TXMLNode.Create(AttribName, AttribValue));
                    State := LookingForAttribNameOrEnd;
                  end
                else
                  AttribValue := AttribValue + '"';
              '''':
                if not DoubleQuotes then
                  begin
                    // Add AttribName, AttribValue
                    If MyChildren = NIL then
                      MyChildren := TList.Create;
                    MyChildren.Add(TXMLNode.Create(AttribName, AttribValue));
                    State := LookingForAttribNameOrEnd;
                  end
                else
                  AttribValue := AttribValue + '''';
            else
              AttribValue := AttribValue + C;
            end;
          LookingForEnd:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              '>': Exit;
              else raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
        end;
      end;
    end;
end;

constructor TXMLNode.CreateRoot(AStream: TStream);
begin
  inherited Create;
  MyTagName := 'Root';
  LoadInternals('', AStream);
end;

destructor TXMLNode.Destroy;
var
  I: Integer;
  Node : TXMLNode;
begin
  if MyChildren <> Nil then
    try
      for I := 0 to MyChildren.Count-1 do
        begin
          Node := TXMLNode(MyChildren[I]);
          MyChildren[I] := Nil;
          Node.Free;
        end;
      MyChildren.Free;
      inherited Destroy;
    except
      // No exceptions out of destructor
    end;
end;

procedure TXMLNode.LoadInternals(IntTagName: String; AStream: TStream);
type
  TState = (
    LookingForTagOrData,
    LookingForTagName,
    ReadingTagName,
    LookingForEndTagName,
    ReadingEndTagName,
    HaveEndTag,
    InComment);
var
  State: TState;
  LineNum: Integer;
  UC, C: Char;
  TagName, EndTagName: String;
begin
  State := LookingForTagOrData;
  LineNum := 1;
  while True do
    begin
      // Read the next character
      if(AStream.Read(UC,1) <> 1) then
        C := #00
      else
        C := UpCase(UC);

      case C of
        #10: Inc(LineNum);
        #13: begin end; // Do nothing
      else
        case State of
          LookingForTagOrData:
            case C of
              #00:
                begin
                  if IntTagName = '' then
                    Exit
                  else
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
                end;
              '<':
                State := LookingForTagName;
              ' ':
                begin
                  if (Length(MyData) > 0) and (MyData[Length(MyData)-1] <> ' ') then
                    MyData := MyData + ' ';
                end
              else
                MyData := MyData + UC;
              end;
          LookingForTagName:
            // Have < so we're looking for / or text
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              ' ': raise Exception.CreateFmt(ErrorUnexpectedWhitespace, [LineNum]);
              '/': State := LookingForEndTagName;
              '!':
                begin
                  if(AStream.Read(C,1) <> 1) then
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum])
                  else if(C <> '-') then
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
                  if(AStream.Read(C,1) <> 1) then
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum])
                  else if(C <> '-') then
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
                  State := InComment;
                end;
              'A'..'Z','0'..'9','_':
                begin
                  AStream.Seek(-1, soFromCurrent);
                  State := ReadingTagName;
                  TagName := '';
                end;
              else raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
          InComment:
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              '-':
                begin
                  if(AStream.Read(C,1) <> 1) then
                    raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum])
                  else if(C = '-') then
                    begin
                      if(AStream.Read(C,1) <> 1) then
                        raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum])
                      else if(C = '>') then
                        State := LookingForTagOrData;
                    end;
                end;
            end;
          ReadingTagName:
            // In the process of loading characters into the end tag name.
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              'A'..'Z','0'..'9','_':
                TagName := TagName + C;
              else
                begin
                  AStream.Seek(-1, soFromCurrent);

                  //TODO: TagName is set...
                  If MyChildren = NIL then
                    MyChildren := TList.Create;
                  MyChildren.Add(TXMLNode.Create(TagName, AStream, LineNum));
                  State := LookingForTagOrData;
                end;
            end;
          LookingForEndTagName:
            // Have </ so we're expecting the end tag name.
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              ' ': raise Exception.CreateFmt(ErrorUnexpectedWhitespace, [LineNum]);
              'A'..'Z','0'..'9','_':
                begin
                  AStream.Seek(-1, soFromCurrent);
                  State := ReadingEndTagName;
                  EndTagName := '';
                end;
              else raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
          ReadingEndTagName:
            // In the process of loading characters into the end tag name.
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              'A'..'Z','0'..'9','_':
                EndTagName := EndTagName + C;
              ' ', '>':
                begin
                  AStream.Seek(-1, soFromCurrent);
                  State := HaveEndTag;
                end;
              else
                raise Exception.CreateFmt(ErrorUnexpectedChar, [C, LineNum]);
            end;
          HaveEndTag:
            // After EndTagName has data.  We're looking for the ending
            // greater-than symbol.
            case C of
              #00: raise Exception.CreateFmt(ErrorUnexpectedEOF, [LineNum]);
              ' ': begin end; //Ignore
              '>':
                begin
                  if(IntTagName <> EndTagName) then
                    raise Exception.CreateFmt(ErrorTagMismatch, [LineNum, EndTagName, IntTagName]);
                  Exit;
                end;
            end;
        end;
      end;
    end;
end;

function TXMLNode.GetItem(AIndex: Integer): TXMLNode;
begin
  Result := TXMLNode(MyChildren[AIndex]);
end;

function TXMLNode.GetCount: Integer;
begin
  Result := MyChildren.Count;
end;

function TXMLNode.GetChildNode(ATagName: String) : TXMLNode;
var
  I : Integer;
begin
  if MyChildren = nil then
    raise Exception.CreateFmt(ErrorNodeNotFound, [ATagName]);
  I := 0;
  while(I < MyChildren.Count) and (ATagName <> TXMLNode(MyChildren[I]).TagName) do
    Inc(I);

  if I >= MyChildren.Count then
    raise Exception.CreateFmt(ErrorNodeNotFound, [ATagName]);

  Result := TXMLNode(MyChildren[I]);
end;

function TXMLNode.GetNode(AStream: TStream) : TXMLNode;
var
  C : Char;
  NodeName : String;
  ChildNode : TXMLNode;
begin
  while True do
    begin
      // Read the next character
      if(AStream.Read(C,1) <> 1) then
        C := #00
      else
        C := UpCase(C);

      case C of
        #00:
          if NodeName = '' then
            begin
              Result := self;
              Exit;
            end
          else
            begin
              Result := GetChildNode(NodeName);
              Exit;
            end;
        'A'..'Z','0'..'9','_':
          NodeName := NodeName + C;
        '/':
          begin
            ChildNode := GetChildNode(NodeName);
            Result := ChildNode.GetNode(AStream);
            Exit;
          end;
      end;
    end
end;

function TXMLNode.GetAttributeValue(AName: String) : String;
var
  I : Integer;
begin
  Result := '';
  if MyChildren = nil then
    Exit;
  I := 0;
  while(I < MyChildren.Count) and (AName <> TXMLNode(MyChildren[I]).TagName) do
    Inc(I);

  if I >= MyChildren.Count then
    Exit;

  Result := TXMLNode(MyChildren[I]).Data;
end;

constructor TXMLParser.Create(const AFileName: String);
var
  Stream: TMemoryStream;
begin
  inherited Create;
  MyFileName := AFileName;
  Stream := TMemoryStream.Create;
  try
    Stream.LoadFromFile(AFileName);
    Stream.Seek(0, soFromBeginning);
    MyRoot := TXMLNode.CreateRoot(Stream);
  finally
    Stream.Free;
  end;
end;

destructor TXMLParser.Destroy;
begin
  MyRoot.Free;
  inherited Destroy;
end;

function TXMLParser.GetNode(Name: String) : TXMLNode;
var
  Stream : TStringStream;
begin
  Stream := TStringStream.Create(Name);
  try
    Stream.Seek(0, soFromBeginning);
    Result := MyRoot.GetNode(Stream);
  finally
    Stream.Free;
  end;
end;

end.
